#include "Drive.h"
#include "../RobotMap.h"
#include "../Commands/AllCommands.h"
#include <math.h>

Drive::Drive() : Subsystem("Drive") {						// Create objects and initialize variables
	f_reverseMaxRate = 15.0;								// Inches per second

	#if (MY_ROBOT == 0)										// Schumacher
		m_motorLF = new VictorSP(PWM_DRIVE_LF);
		m_motorLR = new VictorSP(PWM_DRIVE_LR);
		m_motorRF = new VictorSP(PWM_DRIVE_RF);
		m_motorRR = new VictorSP(PWM_DRIVE_RR);
	#else													// Otis
		m_motorLF = new Victor(PWM_DRIVE_LF);
		m_motorLR = new Victor(PWM_DRIVE_LR);
		m_motorRF = new Victor(PWM_DRIVE_RF);
		m_motorRR = new Victor(PWM_DRIVE_RR);
	#endif

	m_gyro = new Gyro(AI_ROTATE_GYRO);
	m_gyro->SetSensitivity(0.007);			//  0.0069  0.007

	m_leftEncoder = new Encoder(DI_ENCODER_LEFT_A, DI_ENCODER_LEFT_B, true);
	m_leftEncoder->SetDistancePerPulse(0.075);

	m_rightEncoder = new Encoder(DI_ENCODER_RIGHT_A, DI_ENCODER_RIGHT_B);
	m_rightEncoder->SetDistancePerPulse(0.075);

//	m_leftSonar =  new Ultrasonic(7, 6, Ultrasonic::kInches);   // Otis
//	m_rightSonar = new Ultrasonic(9, 8, Ultrasonic::kInches);
	m_leftSonar =  new Ultrasonic(DI_SONAR_L_PING, DI_SONAR_L_ECHO, Ultrasonic::kInches);
	m_rightSonar = new Ultrasonic(DI_SONAR_R_PING, DI_SONAR_R_ECHO, Ultrasonic::kInches);

	m_drivePID = new PIDControl();

	m_rotatePID = new PIDControl();
	SetRotatePID();

	m_atTrigger = m_onTarget = m_rampDone = false;

	m_autoMode = aNone;
	m_driveMode = mDriveRotate;
	m_angleMark = m_angleFeeder = m_angleTarget = 0;
	m_distLast = m_distTarget = m_distTrigger = 0;
	m_maxDrive = m_maxRotate = m_maxStrafe = 0;
	m_pwmDrive = m_pwmRotate = m_pwmStrafe = 0;
	m_rampDrive = 0;
	m_autoCounter = m_doneCounter = 0;
	m_useEncoder = eBoth;
}

Drive::~Drive() {
	delete m_motorLF;
	delete m_motorLR;
	delete m_motorRF;
	delete m_motorRR;
	delete m_gyro;
	delete m_leftEncoder;
	delete m_rightEncoder;
	delete m_leftSonar;
	delete m_rightSonar;
	delete m_drivePID;
	delete m_rotatePID;
}

void Drive::CancelCommand() {
	m_autoMode = aNone;
	m_driveMode = mDriveRotate;

	m_motorLF->Set(0);
	m_motorLR->Set(0);
	m_motorRF->Set(0);
	m_motorRR->Set(0);
}

void Drive::Execute(bool showPID) {
	float 		error;
	static bool feederRotate = false;
	static bool inReverse = false;

	if (m_autoMode == aNone) {
		if (MyRobot::oi->GetFeederRotateButton()) {
			if (!feederRotate) {
				feederRotate = true;
				m_rotatePID->SetOutputRange(-0.7, 0.7);
				m_rotatePID->Reset();
				sprintf(m_log, "Drive:    Feeder Rotate Init: Angle=%5.1f  Gyro=%5.1f", m_angleFeeder, GetGyroAngle());
				MyRobot::robotLog->Write(m_log);
			}

			m_pwmRotate = m_rotatePID->Calculate(RotateError(m_angleFeeder), false);

		} else if (feederRotate) {
			feederRotate = false;
		}

		if (MyRobot::oi->GetFeederDriveButton()) {
			if (m_driveMode != mFeederDrive) {
				m_driveMode = mFeederDrive;

				m_distTarget = (m_leftSonar->GetRangeInches() + m_rightSonar->GetRangeInches()) / 2 - 10.5;
				SetDrivePID((float)(m_distTarget * 0.2));
				m_drivePID->SetOutputRange(-0.6, 0.6);
				m_drivePID->SetSetpoint(m_distTarget);
				m_drivePID->Reset();
				m_pwmDrive  = m_rampDrive = 0;

				m_rotatePID->SetOutputRange(-0.7, 0.7);
				m_rotatePID->Reset();

				m_leftEncoder->Reset();
				m_rightEncoder->Reset();
				sprintf(m_log, "Drive:    Feeder Drive Init: Distance=%5.1f  Angle=%5.1f  Gyro=%5.1f", m_distTarget, m_angleFeeder, GetGyroAngle());
				MyRobot::robotLog->Write(m_log);
			}

		} else if (m_driveMode == mFeederDrive) {
			m_driveMode = mDriveRotate;
		}

		switch (m_driveMode) {
			case mDriveRotate:
				m_pwmDrive = MyRobot::oi->GetDrive();
				m_pwmStrafe = 0;
				if (!feederRotate) m_pwmRotate = MyRobot::oi->GetRotate();
				break;

			case mDriveStrafe:
				m_pwmDrive = MyRobot::oi->GetDrive();
				m_pwmStrafe = MyRobot::oi->GetStrafe();
				if (!feederRotate) m_pwmRotate = m_rotatePID->Calculate(RotateError(m_angleTarget), false);
				break;

			case mStrafe:
				m_pwmDrive = 0;
				m_pwmStrafe = MyRobot::oi->GetStrafe();
				if (!feederRotate) m_pwmRotate = m_rotatePID->Calculate(RotateError(m_angleTarget), false);
				break;

			case mFeederDrive:
				m_pwmDrive = DrivePWM(m_pwmDrive, m_drivePID->Calculate(GetEncoderDistance(eBoth), false));
				m_pwmStrafe = 0;
				m_pwmRotate = m_rotatePID->Calculate(RotateError(m_angleFeeder), false);
				break;
			default:;
		}

		if (!inReverse && m_pwmDrive < 0) {
			if (MyRobot::elevator->IsInCarry()) {
				if (m_leftEncoder->GetRate() > f_reverseMaxRate) {
					m_pwmDrive = 0;
				}
			}
		}

		inReverse = (m_pwmDrive < 0);

	} else if (m_autoMode == aRotate) {
		error = RotateError(m_angleTarget);					// Determine error from target

		m_pwmDrive = m_pwmStrafe = 0;
		m_pwmRotate = m_rotatePID->Calculate(error, showPID);	// Get rotate speed from PID

		if (fabs(error) <= 1.0 ) {							// Rotation with 1 degree of target
			if (m_doneCounter < 3){
				m_doneCounter++;
			} else {										// Within 1 degree of target for at least 60 ms
				m_pwmRotate = 0;

				if(!m_onTarget) {
					m_onTarget = true;
					m_autoMode = aNone;
					sprintf(m_log, "Drive:    Rotate Done: Angle=%5.1f", GetGyroAngle());
					MyRobot::robotLog->Write(m_log);
				}
			}
		} else {
			m_doneCounter = 0;
		}

	} else {
		float curDistance = GetEncoderDistance(m_useEncoder);	// Get current distance
		error = RotateError(m_angleTarget);						// Determine error from target

		switch(m_autoMode) {
			case aDrive:
				m_pwmDrive = DrivePWM(m_pwmDrive, m_drivePID->Calculate(curDistance, showPID));
				m_pwmStrafe = 0;
				m_pwmRotate = m_rotatePID->Calculate(error, false);
				break;

			case aStrafe:
				m_pwmDrive = 0;
				m_pwmRotate = m_rotatePID->Calculate(error, false);

				if (fabs(curDistance) > fabs(m_distTarget)) {
					m_pwmStrafe = 0;
				} else {
					m_pwmStrafe = m_maxStrafe;
				}
				break;

			case aDriveTurn:
				m_pwmDrive = DrivePWM(m_pwmDrive, m_drivePID->Calculate(curDistance, showPID));
				m_pwmStrafe = 0;

				if (fabs(error) <= 5.0) {
					m_autoMode = aDrive;
					m_pwmRotate = m_rotatePID->Calculate(error, false);
				} else {
					m_pwmRotate = m_maxRotate * fabs(m_pwmDrive / m_maxDrive);
				}
				break;

			case aDriveStrafe:
				m_pwmDrive = DrivePWM(m_pwmDrive, m_drivePID->Calculate(curDistance, showPID));
				m_pwmStrafe = m_maxStrafe * fabs(m_pwmDrive / m_maxDrive);
				m_pwmRotate = m_rotatePID->Calculate(error, false);
				break;

			default:;
		}

		curDistance = fabs(curDistance);

		if (m_distTrigger != 0) m_atTrigger = (curDistance >= m_distTrigger);

		if (curDistance > fabs(m_distTarget * 0.5)) {			// More than half way to target
			if (curDistance >= fabs(m_distTarget)) {
				if (m_distLast < fabs(m_distTarget)) {
					sprintf(m_log, "Drive:    Drive at Target: Distance=%6.1f  Change=%6.1f", GetEncoderDistance(m_useEncoder),
							curDistance - m_distLast);
					MyRobot::robotLog->Write(m_log);
				}
			}

//			if ((curDistance - m_distLast <= 0) || (curDistance >= fabs(m_distTarget))) {
			if (curDistance - m_distLast <= 0) {				// Robot has stopped or is moving in opposite direction
				m_pwmDrive = 0;
				m_pwmStrafe = 0;

				if (fabs(error) <= 1.0 ) {						// Rotation is within 1 degree of heading
					m_pwmRotate = 0;
					m_onTarget = true;
					m_autoMode = aNone;
					sprintf(m_log, "Drive:    Drive Done: Distance=%6.1f  Angle=%5.1f", GetEncoderDistance(m_useEncoder), GetGyroAngle());
					MyRobot::robotLog->Write(m_log);
				}
			}
		}

		m_distLast = curDistance;								// Save current distance for next execution
	}

	MecanumDrive(m_pwmDrive, m_pwmStrafe, m_pwmRotate, wAll);
}

double Drive::GetEncoderDistance(DriveEncoder encoder){
	double left = 	m_leftEncoder->GetDistance();
	double right = 	m_rightEncoder->GetDistance();

	switch(encoder) {
		case eLeft:		return left;
		case eRight:	return right;
		case eBoth:		if (fabs(left)  < fabs(right * 0.5)) 	return right;
						if (fabs(right) < fabs(left * 0.5)) 	return left;
						if (fabs(left)  < fabs(right)) 			return left;
																return right;
		default:		return 0;
	}
}

float Drive::GetGyroAngle() {								// Determine angle of robot in range of 0 to 360 degrees
	float angle = ((int)(m_gyro->GetAngle() * 10) % 3600) / 10.0;
	if (angle < 0) angle += 360;
	return angle;
}

void Drive::InitAutoDrive(bool resetEncoder, DriveEncoder useEncoder, bool useBrake, float angle, AngleFrom angleFrom,
						  double target, double trigger, float driveMax, float driveRamp){
	InitAuto(resetEncoder, useEncoder, useBrake, angle, angleFrom, target, trigger);

	m_autoMode = aDrive;
	m_maxDrive = driveMax;
	m_rampDrive = driveRamp;
	m_drivePID->SetOutputRange(-m_maxDrive, m_maxDrive);
	m_rotatePID->SetOutputRange(-0.7, 0.7);

	sprintf(m_log, "Drive:    Auto Drive Init: Distance=%6.1f  Heading=%5.1f", m_distTarget, m_angleTarget);
	MyRobot::robotLog->Write(m_log);
}

void Drive::InitAutoDriveStrafe(bool resetEncoder, DriveEncoder useEncoder,bool useBrake, float angle, AngleFrom angleFrom,
								double target, double trigger, float driveMax, float driveRamp, float strafeMax) {
	InitAuto(resetEncoder, useEncoder, useBrake, angle, angleFrom, target, trigger);

	m_maxDrive = driveMax;
	m_maxStrafe = strafeMax;
	m_rampDrive = driveRamp;
	m_rotatePID->SetOutputRange(-0.7, 0.7);

	m_autoMode = aDriveStrafe;
	m_drivePID->SetOutputRange(-m_maxDrive, m_maxDrive);
	sprintf(m_log, "Drive:    Auto DriveStrafe Init: Distance=%6.1f  Heading=%5.1f", m_distTarget, m_angleTarget);
	MyRobot::robotLog->Write(m_log);
}

void Drive::InitAutoDriveTurn(bool resetEncoder, DriveEncoder useEncoder,bool useBrake, float angle, AngleFrom angleFrom,
							  double target, double trigger, float driveMax, float driveRamp, float rotateMax) {
	InitAuto(resetEncoder, useEncoder, useBrake, angle, angleFrom, target, trigger);

	m_autoMode = aDriveTurn;
	m_maxDrive = driveMax;
	m_rampDrive = driveRamp;
	m_maxRotate = rotateMax;
	m_drivePID->SetOutputRange(-m_maxDrive, m_maxDrive);
	m_rotatePID->SetOutputRange(-fabs(m_maxRotate), fabs(m_maxRotate));

	sprintf(m_log, "Drive:    Auto Turn Init: Distance=%6.1f  Heading=%5.1f", m_distTarget, m_angleTarget);
	MyRobot::robotLog->Write(m_log);
}

void Drive::InitAutoRotate(float angle, AngleFrom angleFrom, float rotateMax) {
	m_autoMode = aRotate;

	if (angleFrom == fFeeder) {
		angle = m_angleFeeder;
	} else if (angleFrom == fNewMark) {						// Determine angle relative to robot start position
		m_angleMark = GetGyroAngle();
		angle += m_angleMark;
	} else if (angleFrom == fMark) {
		angle += m_angleMark;
	}

	m_angleTarget = ((int)(angle * 10) % 3600) / 10.0;		// Limit angle to range of 0 to 360 degrees
	if (m_angleTarget < 0) m_angleTarget += 360;

	m_rotatePID->SetOutputRange(-rotateMax, rotateMax);
	m_rotatePID->Reset();									// Reset Rotate PID
	m_pwmRotate = 0;
	m_onTarget = false;

	sprintf(m_log, "Drive:    Auto Rotate Init: From=%5.1f  To=%5.1f", GetGyroAngle(), m_angleTarget);
	MyRobot::robotLog->Write(m_log);
}

void Drive::InitAutoStrafe(bool resetEncoder, DriveEncoder useEncoder,bool useBrake, float angle, AngleFrom angleFrom,
						   double target, double trigger, float strafeMax) {
	InitAuto(resetEncoder, useEncoder, useBrake, angle, angleFrom, target, trigger);

	m_autoMode = aStrafe;
	m_maxStrafe = strafeMax;
	m_rotatePID->SetOutputRange(-0.7, 0.7);

	sprintf(m_log, "Drive:    Auto Strafe Init: Distance=%6.1f  Heading=%5.1f", m_distTarget, m_angleTarget);
	MyRobot::robotLog->Write(m_log);
}

void Drive::InitDefaultCommand()
{
//	SetDefaultCommand(new DriveJoystick());
}

bool Drive::IsAtTrigger() {
	return m_atTrigger;
}

bool Drive::IsOnTarget() {									// Drive and/or Rotate positions are at target
	return m_onTarget;
}

void Drive::ResetEncoders() {
	m_leftEncoder->Reset();
	m_rightEncoder->Reset();
}

void Drive::SetConstant(std::string key, int32_t value) {
	if(key == "reverseMaxRate") {
		f_reverseMaxRate = (double)value / 10;
		sprintf(m_log, "Drive:    Set Reverse Max Rate=%4.1f", f_reverseMaxRate);
		MyRobot::robotLog->Write(m_log);
	}
}

void Drive::SetDriveMode(DriveMode mode) {
	m_driveMode = mode;

	if (mode == mDriveStrafe || mode == mStrafe) {
		m_angleTarget = GetGyroAngle();
		m_rotatePID->Reset();
		m_rotatePID->SetOutputRange(-0.7, 0.7);
	}
}

void Drive::SetDrivePID(float dThreshold) {
	m_drivePID->SetCoefficient('P', 0, 0.02, 0);
	m_drivePID->SetCoefficient('I', 12, 0, 0.001);			// 6, 0, 0.001
	m_drivePID->SetCoefficient('D', dThreshold, 0, 0.3);

	m_drivePID->SetInputRange(-200, 200);
	m_drivePID->SetOutputRange(-0.6, 0.6);
}

void Drive::SetMarkAngle() {
	m_angleMark = GetGyroAngle();
}

bool Drive::SetFeederAngle(bool useSonar, bool logReset) {
	static bool		logDone = false;
	static int32_t 	newAngleCount = 0;
	double 			leftRange =  m_leftSonar->GetRangeInches();
	double 			rightRange = m_rightSonar->GetRangeInches();
	bool   			angleSet = false;

	if (logReset) logDone = false;

	if (useSonar) {
		if (fabs(m_leftEncoder->GetRate()) < 10.0 && fabs(m_rightEncoder->GetRate()) < 10.0) {
			if (leftRange <= 48 && rightRange <= 48) {
				m_angleFeeder = ((int)((GetGyroAngle() + atan2(leftRange - rightRange, 24) * 180 / 3.14159) * 10) % 3600) / 10.0;
				if (m_angleFeeder < 0) m_angleFeeder += 360;

				newAngleCount = (newAngleCount == 1) ? 2 : 1;
				MyRobot::dashboard->SetRobotValue(RV_NEW_FEEDER_ANGLE, newAngleCount);

				angleSet = true;
				sprintf(m_log, "Drive:    Set Feeder Angle w/Sonar: L=%5.1f  R=%5.1f  Angle=%5.1f", leftRange, rightRange, m_angleFeeder);
				MyRobot::robotLog->Write(m_log);

			} else if (!logDone) {
				logDone = true;
				sprintf(m_log, "Drive:    Set Feeder Angle w/Sonar: Out of range");
				MyRobot::robotLog->Write(m_log);
			}
		} else if (!logDone) {
			logDone = true;
			sprintf(m_log, "Drive:    Set Feeder Angle w/Sonar: Robot Moving");
			MyRobot::robotLog->Write(m_log);
		}

	} else {
		m_angleFeeder = GetGyroAngle();
		angleSet = true;
		sprintf(m_log, "Drive:    Set Feeder Angle w/Gyro: Angle=%5.1f", m_angleFeeder);
		MyRobot::robotLog->Write(m_log);
	}

	return angleSet;
}

void Drive::SetSonarEnable(bool enable) {
	m_leftSonar->SetAutomaticMode(enable);
	m_rightSonar->SetAutomaticMode(enable);
}

void Drive::SetRotatePID() {
	m_rotatePID->SetCoefficient('P', 0, 0.03, 0);			//	0.04
	m_rotatePID->SetCoefficient('I', 10, 0, 0.004);
	m_rotatePID->SetCoefficient('D', 0, 0.1, 0);

	m_rotatePID->SetInputRange(-180.0, 180.0);
	m_rotatePID->SetOutputRange(-0.7, 0.7);
	m_rotatePID->SetSetpoint(0);
}

void Drive::TuneDrivePID() {
	m_drivePID->SetCoefficient('P', (float)MyRobot::dashboard->GetDashValue(DV_PID_P_THRESHOLD),
									(float)MyRobot::dashboard->GetDashValue(DV_PID_P_ABOVE) / 10000,
									(float)MyRobot::dashboard->GetDashValue(DV_PID_P_BELOW) / 10000);
	m_drivePID->SetCoefficient('I', (float)MyRobot::dashboard->GetDashValue(DV_PID_I_THRESHOLD),
									(float)MyRobot::dashboard->GetDashValue(DV_PID_I_ABOVE) / 10000,
									(float)MyRobot::dashboard->GetDashValue(DV_PID_I_BELOW) / 10000);
	m_drivePID->SetCoefficient('D', (float)MyRobot::dashboard->GetDashValue(DV_PID_D_THRESHOLD),
									(float)MyRobot::dashboard->GetDashValue(DV_PID_D_ABOVE) / 10000,
									(float)MyRobot::dashboard->GetDashValue(DV_PID_D_BELOW) / 10000);

	InitAutoDrive(true, Drive::eBoth, true, 0, Drive::fNewMark,
				  (double)MyRobot::dashboard->GetDashValue(DV_PID_SETPOINT), 0,
				  (float)MyRobot::dashboard->GetDashValue(DV_PID_MAX_PWM) / 100, 0);

}

void Drive::TuneRotatePID() {
	m_rotatePID->SetCoefficient('P', (float)MyRobot::dashboard->GetDashValue(DV_PID_P_THRESHOLD),
									 (float)MyRobot::dashboard->GetDashValue(DV_PID_P_ABOVE) / 10000,
									 (float)MyRobot::dashboard->GetDashValue(DV_PID_P_BELOW) / 10000);
	m_rotatePID->SetCoefficient('I', (float)MyRobot::dashboard->GetDashValue(DV_PID_I_THRESHOLD),
									 (float)MyRobot::dashboard->GetDashValue(DV_PID_I_ABOVE) / 10000,
									 (float)MyRobot::dashboard->GetDashValue(DV_PID_I_BELOW) / 10000);
	m_rotatePID->SetCoefficient('D', (float)MyRobot::dashboard->GetDashValue(DV_PID_D_THRESHOLD),
									 (float)MyRobot::dashboard->GetDashValue(DV_PID_D_ABOVE) / 10000,
									 (float)MyRobot::dashboard->GetDashValue(DV_PID_D_BELOW) / 10000);

	InitAutoRotate((float)MyRobot::dashboard->GetDashValue(DV_PID_SETPOINT), fNewMark,
			       (float)MyRobot::dashboard->GetDashValue(DV_PID_MAX_PWM) / 100);
}

// ******************** PRIVATE ********************

void Drive::ArcadeDrive (float drive, float rotate) {
	//    drive
	//		|			Joystick Quadrants
	//  1	|	2
	//		|
	// ------------ rotate
	//		|
	//	3	|	4
	//		|

	float leftPwm;
	float rightPwm;

	if (drive > 0.0) {										// Quadrant 1 or 2
		if(rotate > 0.0) { 									// Quadrant 2
			leftPwm = drive - rotate;
			rightPwm = (drive > rotate ? drive : rotate);
		} else { 											// Quadrant 1 or drive axis
			leftPwm = (drive > -rotate ? drive : -rotate);
			rightPwm = drive + rotate;
		}

	} else {												// Quadrant 3, 4, or rotate axis
		if(rotate > 0.0) { 									// Quadrant 4
			leftPwm = -(-drive > rotate ? -drive: rotate);
			rightPwm = drive + rotate;
		} else { 											// Quadrant 3 or drive axis
			leftPwm = drive - rotate;
			rightPwm = -(-drive > -rotate ? -drive : -rotate);
		}
	}

	m_motorLF->Set(leftPwm);
	m_motorLR->Set(leftPwm);
	m_motorRF->Set(-rightPwm);
	m_motorRR->Set(-rightPwm);

	//printf("Drive=%f  Rotate=%f Left=%f Right=%f\n", drive, rotate, leftPwm, -rightPwm);
}

float Drive::DrivePWM(float curPWM, float pidPWM) {
	if (m_rampDone) return pidPWM;

	float direction = (pidPWM < 0 ? -1.0 : 1.0);

	curPWM = fabs(curPWM);
	pidPWM = fabs(pidPWM);									// Use absolute value for PWMs

	if (curPWM == 0.0) {									// If stopped, set speed to minimum value
		curPWM = 0.15;
	} else if(m_rampDrive == 0) {
		curPWM += 0.02;										// 0.02  0.008  0.005
	} else {												// Ramp speed
		curPWM += m_rampDrive;
	}

	if (curPWM >= pidPWM) {
		curPWM = pidPWM;
		m_rampDone = true;
	}

	return curPWM * direction;								// Apply direction and set the current PWM
}

void Drive::InitAuto(bool resetEncoder, DriveEncoder useEncoder, bool useBrake, float angle, AngleFrom angleFrom,
					 double target, double trigger) {
	m_distTarget = target;
	m_distTrigger = fabs(trigger);
	m_useEncoder = useEncoder;

	m_autoCounter = 20;
	m_distLast = m_doneCounter = 0;
	m_maxDrive = m_maxRotate = m_maxStrafe = 0;
	m_pwmDrive = m_pwmRotate = m_pwmStrafe = 0;

	m_atTrigger = m_onTarget = m_rampDone = false;

	if(resetEncoder) {
		m_leftEncoder->Reset();
		m_rightEncoder->Reset();
	}

	if (useBrake) {
		SetDrivePID((float)(m_distTarget * 0.2));
	} else {
		SetDrivePID(0);
	}

	m_drivePID->Reset();									// Configure Drive PID
	m_drivePID->SetSetpoint(m_distTarget);

	if (angleFrom == fFeeder) {
		angle = m_angleFeeder;
	} else if (angleFrom == fNewMark) {						// Determine heading and reset Rotate PID
		m_angleMark = GetGyroAngle();
		angle += m_angleMark;
	} else if (angleFrom == fMark) {
		angle += m_angleMark;
	}

	m_angleTarget = ((int)(angle * 10) % 3600) / 10.0;		// Limit heading to range of 0 to 360 degrees
	if (m_angleTarget < 0) m_angleTarget += 360;

	m_rotatePID->Reset();
}

float Drive::Limit(float Value){							// Limit Value to range of -1.0 to +1.0
	if(Value > 1.0)  return 1.0;
	if(Value < -1.0) return -1.0;
					 return Value;
}

void Drive::MecanumDrive(float drive, float strafe, float rotate, DriveWheels driveWheels){
	float MotorAbs;
	float MotorPwm[4];
	float MotorMax = 0;

	MotorPwm[0] = rotate + drive + strafe;
	MotorPwm[1] = rotate + drive - strafe;
	MotorPwm[2] = rotate - drive + strafe;
	MotorPwm[3] = rotate - drive - strafe;

	for(int i = 0; i < 4; i++) {
		MotorAbs = fabs(MotorPwm[i]);
		if(MotorMax < MotorAbs) MotorMax = MotorAbs;
	}

	for(int i = 0; i < 4; i++) {
		if(MotorMax > 1.0) MotorPwm[i] = Limit(MotorPwm[i] / MotorMax);
	}

	switch(driveWheels) {
		case wAll:
			m_motorLF->Set(MotorPwm[0]);
			m_motorLR->Set(MotorPwm[1]);
			m_motorRF->Set(MotorPwm[2]);
			m_motorRR->Set(MotorPwm[3]);
			break;
		case wRear:
			m_motorLF->Set(0);
			m_motorLR->Set(MotorPwm[1]);
			m_motorRF->Set(0);
			m_motorRR->Set(MotorPwm[3]);
			break;
		case wFront:
			m_motorLF->Set(MotorPwm[0]);
			m_motorLR->Set(0);
			m_motorRF->Set(MotorPwm[2]);
			m_motorRR->Set(0);
			break;
		default:;
	}

	//printf("LF = %f  LR = %f  RF = %f  RR = %f\n", MotorPwm[0], MotorPwm[1], MotorPwm[2], MotorPwm[3]);
}

float Drive::RotateError(float target) {					// Calculate rotation error in -180 to +180 range
	float error = GetGyroAngle() - target;

	if (error > 180) {
		error -= 360;
	} else if (error < -180) {
		error += 360;
	}

//	printf("Rotate Error: target=%f  error=%f\n", target, error);
	return error;
}

